@@ -41,11 +41,16 @@ def int_cast(s: str) -> Optional[int]:
     return int(s, base=0) if s is not None else None
 
 
+# special "path" type
+class path(str):
+    pass
+
 # note: mypy complains wrongly about non-existing _empty.
 CASTS: Dict[type, Callable[[str], object]] = {
     bool: bool_cast,
     int: int_cast,
     inspect._empty: str,
+    path: str,
     dt.date: dt.date.fromisoformat,
     dt.time: dt.time.fromisoformat,
     dt.datetime: dt.datetime.fromisoformat
@@ -71,6 +76,75 @@ def typeof(p: inspect.Parameter):
         return str
 
 
+# convenient object wrapper class
+class Reference:  # type: Any
+
+    def __init__(self, obj: Any = None):
+        self._obj = None
+        # keep track of initial methods
+        self._init = set(self.__dir__())
+        self._init.add("_init")
+        if obj is not None:
+            self._setobj(obj)
+
+    def _setobj(self, obj):
+        log.debug(f"setting reference to {obj} ({type(obj)})")
+        self._obj = obj
+        # method cleanup
+        for f in self.__dir__():
+            if f not in self._init:
+                delattr(self, f)
+        # forward
+        for f in obj.__dir__():
+            if f not in self._init:
+                setattr(self, f, getattr(obj, f))
+        # and more forwards
+        # for f in ("str", "repr", "hash", "eq", "ne", "lt", "gt", "le", "ge"):
+        #    name = "__" + f + "__"
+        #    setattr(self, name, getattr(obj, name))
+
+    def __str__(self):
+        return self._obj.__str__()
+
+    def __repr__(self):
+        return self._obj.__repr__()
+
+    def __eq__(self, *args):
+        return self._obj.__eq__(*args)
+
+    def __hash__(self):
+        return self._obj.__hash__()
+
+
+#
+# Positive caching.
+#
+# Keep a cache of true answers to reduce database queries,
+# but keep querying on failures.
+#
+# See also: functools.cache, functools.lru_cache
+#
+# TODO
+# - LRU? LFU?
+# - automatic reset based on cache efficiency? expansion?
+#
+class CacheOK:
+
+    def __init__(self, fun):
+        self._fun = fun
+        self._cache = set()
+        self.cache_clear = self._cache.clear
+
+    def __call__(self, *args):
+        if args in self._cache:
+            return True
+        else:
+            ok = self._fun(*args)
+            if ok:
+                self._cache.add(args)
+            return ok
+
+
 # Flask wrapper
 class Flask(flask.Flask):
 
@@ -103,6 +177,9 @@ class Flask(flask.Flask):
     def get_user(self):
         return self._fsa.get_user()
 
+    def current_user(self):
+        return self._fsa._user
+
 
 # actual class
 class FlaskSimpleAuth:
@@ -164,6 +241,7 @@ class FlaskSimpleAuth:
         # auth setup
         #
         self._auth = conf.get("FSA_TYPE", "httpd")
+        assert self._auth in ("httpd", "none", "fake", "basic", "param", "password", "token")
         self._lazy = conf.get("FSA_LAZY", True)
         self._always = conf.get("FSA_ALWAYS", True)
         self._check = conf.get("FSA_CHECK", True)
@@ -417,12 +495,16 @@ class FlaskSimpleAuth:
         # all is well
         return user
 
-    def _get_jwt_token_auth(self, token):
+    # jwt authentication can be expensive, especially with pubkey-signatures
+    # so use a cache to keep track of already used tokens
+    @functools.lru_cache(maxsize=1024)
+    def _get_jwt_token_auth_real(self, token):
         import jwt
         try:
             data = jwt.decode(token, self._secret, leeway=self._delay * 60,
                               audience=self._realm, algorithms=[self._algo])
-            return data['sub']
+            exp = dt.datetime.fromtimestamp(data['exp'])
+            return data['sub'], exp
         except jwt.ExpiredSignatureError:
             log.debug(f"LOGIN (token): token {token} has expired")
             raise AuthException("expired jwt auth token", 401)
@@ -430,6 +512,15 @@ class FlaskSimpleAuth:
             log.debug(f"LOGIN (token): invalide token ({e})")
             raise AuthException("invalid jwt token", 401)
 
+    def _get_jwt_token_auth(self, token):
+        user, exp = self._get_jwt_token_auth_real(token)
+        # recheck token expiration
+        now = dt.datetime.utcnow() - dt.timedelta(minutes=self._grace)
+        if now > exp:
+            log.debug(f"LOGIN (token): token {token} has expired")
+            raise AuthException("expired jwt auth token", 401)
+        return user
+
     def _get_token_auth(self, token):
         log.debug(f"checking token: {token}")
         return \
@@ -439,14 +530,16 @@ class FlaskSimpleAuth:
     def _get_password_auth(self):
         try:
             return self._get_basic_auth()
-        except AuthException:  # try param
+        except AuthException:  # failed, let's try param
             return self._get_param_auth()
 
     # map auth types to their functions
-    _FSA_AUTH = {"basic": _get_basic_auth,
-                 "param": _get_param_auth,
-                 "password": _get_password_auth,
-                 "fake": _get_fake_auth}
+    _FSA_AUTH = {
+        "basic": _get_basic_auth,
+        "param": _get_param_auth,
+        "password": _get_password_auth,
+        "fake": _get_fake_auth
+    }
 
     # return authenticated user or throw exception
     def get_user(self):
@@ -457,15 +550,17 @@ class FlaskSimpleAuth:
         if self._user is not None:
             return self._user
 
-        AUTH = self._auth
-        if AUTH is None:
+        a = self._auth
+        if a is None:
             raise AuthException("FlaskSimpleAuth not initialized", 500)
 
-        if AUTH == "httpd":
+        elif a == "none":
+            return None
 
+        elif a == "httpd":
             self._user = request.remote_user
 
-        elif AUTH in ("fake", "param", "basic", "token", "password"):
+        elif a in ("fake", "basic", "param", "password", "token"):
 
             # check for token
             if self._type is not None:
@@ -481,13 +576,13 @@ class FlaskSimpleAuth:
 
             # else try other schemes
             if self._user is None:
-                if AUTH in self._FSA_AUTH:
-                    self._user = self._FSA_AUTH[AUTH](self)
+                if a in self._FSA_AUTH:
+                    self._user = self._FSA_AUTH[a](self)
                 else:
                     raise AuthException("auth token is required", 401)
 
         else:
-            raise AuthException(f"unexpected authentication type: {AUTH}", 500)
+            raise AuthException(f"unexpected authentication type: {a}", 500)
 
         assert self._user is not None  # else an exception would have been raised
         log.info(f"get_user({self._auth}): {self._user}")
@@ -658,7 +753,7 @@ class FlaskSimpleAuth:
                     t = typeof(sig.parameters[spec])
                     # Flask supports 5 types, with string the default?
                     # FIXME how to handle path?
-                    if t in (int, float, UUID):
+                    if t in (int, float, UUID, path):
                         splits[i] = f"{t.__name__.lower()}:{spec}>{remainder}"
                     else:
                         splits[i] = f"string:{spec}>{remainder}"
